import pandas as pd
import lzma
import numpy as np
from Bio import SeqIO
import os
import re
import unicodedata
import tarfile
import io
import gzip


def clean_string(text, strip_symbols=False):
    # strip accents
    text = ''.join((c for c in unicodedata.normalize('NFD', text)
                    if unicodedata.category(c) != 'Mn'))
    # replace spaces with _
    text = text.replace(' ', '_')

    if strip_symbols:
        p = re.compile(r'[^a-zA-Z0-9_/-]')
        text = p.sub('', text)
    return text


def fix_header(header):
    """
    parse fasta header and remove problems
    """
    fixed_header = header.lstrip()\
        .replace(' ', '_')\
        .replace("hCoV-19/","")\
        .replace("hCov-19/","")\
        .replace("PENDING", "")\
        .replace("UPLOADED", "")\
        .replace("None", "")

    fixed_header = re.sub('^Wuhan-Hu-1', 'China/Wuhan-Hu-1', fixed_header)
    fixed_header = clean_string(fixed_header)

    return (fixed_header)

# Increase number of rows to print for pandas
pd.set_option('display.max_rows', 50)
pd.set_option('display.min_rows', 50)

#Create argparser
import argparse
parser = argparse.ArgumentParser(description='Create input for Armadillo training')
parser.add_argument('--designations', type=str, help='Path to pango-designation repository')
parser.add_argument('--output', type=str, help='Path to output directory')
parser.add_argument('--gisaid_meta_file', type=str, help='Path to GISAID metadata file')
parser.add_argument('--gisaid_mmsa', type=str, help='Path to GISAID MSA')


args = parser.parse_args()

#If output dir does not exist, create it:
if not os.path.exists(args.output):
    os.makedirs(args.output)


assignments = pd.read_csv(f"{args.designations}/lineages.csv")
all_lineages = sorted(list(set(assignments['lineage'])))
# save all_lineages to file
with open(f"{args.output}/all_lineages.txt", "w") as f:
    for l in all_lineages:
        f.write(f"{l}\n")

# read entire contents of aliases file
aliases_file = open(f"{args.designations}/pango_designation/alias_key.json", "rt")

# save contents of aliases to output
with open(f"{args.output}/aliases.json", "w") as f:
    for line in aliases_file:
        f.write(line)



total_sampling = 1

lineage_counts = assignments['lineage'].value_counts().to_frame()

print(f"AY.4 starting count is {lineage_counts['lineage']['AY.4']}")
print(f"AY.6 starting count is {lineage_counts['lineage']['AY.6']}")

print(f"B.1.1.7 starting count is {lineage_counts['lineage']['B.1.1.7']}")
print(f"Q.1 starting count is {lineage_counts['lineage']['Q.1']}")


lineage_counts['nonlinear'] =lineage_counts['lineage']**.22
lineage_counts['nonlinear_ratio'] =lineage_counts['nonlinear'].sum()/ lineage_counts['nonlinear'] 
lineage_counts['nonlinear_ratio'] = lineage_counts['nonlinear_ratio']  / 500
lineage_counts['expectation'] = lineage_counts['lineage'] * lineage_counts['nonlinear_ratio']



print(f"Expecting total size of: {lineage_counts['expectation'].sum()}")
print(f"Expecting this many Q.1: {lineage_counts['expectation']['Q.1']}")
print(f"Expecting this many B.1.1.7: {lineage_counts['expectation']['B.1.1.7']}")




metadata = pd.read_csv(args.gisaid_meta_file,
                       sep="\t",
                       usecols=["Virus name", "Accession ID"])

# rename Virus name to strain and Accession ID to gisaid_epi_isl
metadata = metadata.rename(columns={"Virus name": "strain", "Accession ID": "gisaid_epi_isl"})
#Replace spaces with underscores in strain
metadata['strain'] = metadata['strain'].apply(fix_header)

epi_to_name = dict(zip(metadata["gisaid_epi_isl"], metadata["strain"]))

name_to_lineage = dict(zip(assignments['taxon'], assignments['lineage']))
#print(name_o_taxon)

# Get list of names that are "AY.6"
AY_6_names = assignments[assignments['lineage'] == 'AY.6']['taxon'].tolist()

#raise ValueError(AY_6_names)
# Get epis from these names
AY_6_epis = metadata[metadata['strain'].isin(AY_6_names)]['gisaid_epi_isl'].tolist()



#raise ValueError((AY_6_epis))

filename = args.gisaid_mmsa



if filename.endswith("tar.xz"):
            tar = tarfile.open(filename, "r:xz")
            while True:
                member = tar.next()
                if member.size>100000:
                    break
            if member is None:
                raise ValueError("Could not find member")
            
            handle = tar.extractfile(member)
            handle = io.TextIOWrapper(handle) 
            print("Using largest file as handle")
            print("Using tarxz mode")
            
else:
    handle = open(filename,"rt")

number_of_shards = 400
file_handles = {}

def get_shard_path(i):
    return f"{args.output}/shard_{i}.tsv.gz"
for i in range(number_of_shards):
    file_handles[i] = gzip.open(get_shard_path(i), "wt")

import tqdm
import random

metadata_num_rows = metadata.shape[0]

def write_to_random_handle(epi,seq
,lineage):
    #print("writing")
    random_handle = file_handles[random.randint(
        0, number_of_shards - 1)]
    random_handle.write(f"{epi}\t{seq}\t{lineage}\n")

seq_is_good = False

total = 0
import random
if True:
    with tqdm.tqdm(total=metadata_num_rows) as pbar:
        for record in SeqIO.parse(handle, "fasta"):
            pbar.update(1)
            total +=1
            #if total>10000:
            #    break
            epi = record.id
            seq = str(record.seq)
            try:
                lineage = name_to_lineage[epi_to_name[epi]]
            except KeyError:
                if epi not in epi_to_name:
                    print(f"{epi} not in epi_to_name")
                else:
                   #print(f"{epi_to_name[epi]} not in name_to_lineage")
                   pass
                continue
            
            times_to_sample = lineage_counts['nonlinear_ratio'][lineage]
            #print(f"{lineage} is {times_to_sample} times nonlinear")
            if times_to_sample > 1:
                for i in range(int(times_to_sample)):
                    write_to_random_handle(epi, seq, lineage)
            else:
                rand_no = random.random()
                if  rand_no< times_to_sample:
                     write_to_random_handle(epi, seq, lineage)

print("Completed, closing all files")
# Now close all files:
for i in range(number_of_shards):
    file_handles[i].close()

print("Beginning shuffle")
# Now open each in turn, shuffle the lines, and write out again:
for i in tqdm.tqdm(range(number_of_shards)):
    print("opening this file")
    handle = gzip.open(get_shard_path(i), "rt")
    
    lines = handle.readlines()
    random.shuffle(lines)
    handle.close()
    print("loaded this file")
    handle = gzip.open(get_shard_path(i), "wt")
    handle.writelines(lines)
    handle.close()


print("Shuffle complete")





import argparse
import os

# Create the parser
parser = argparse.ArgumentParser(description='Train the model')
parser.add_argument('--starting_model',type=str)
parser.add_argument('--do_pruning', action='store_true')
parser.add_argument('--checkpoint_path',type=str)
parser.add_argument('--disable_gpu', action='store_true')
parser.add_argument('--use_wandb', action='store_true')
parser.add_argument('--shard_dir', type=str)
parser.add_argument('--epochs', type=int, default=None)
args = parser.parse_args()

if args.disable_gpu:
    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

import numpy as np
import tensorflow_model_optimization as tfmot
import pandas as pd
import lzma
from . import modelling
import tensorflow as tf
from . import training_input


# Create a callback to print stuff
class PrintZeros(tf.keras.callbacks.Callback):
    def on_epoch_end(self, epoch, logs=None):
        layer = self.model.get_layer(
        "prune_low_magnitude_multiply_by_weights")
        weights = layer.get_weights()[0]
        num_zeros = np.sum(weights == 0)
        num_non_zeros = np.sum(weights != 0)
        print("Number of zeros:", num_zeros)
        print("Number of non-zeros:", num_non_zeros)
        percent = num_non_zeros / (num_zeros + num_non_zeros)
        print("Proportion of non-zeros:", percent)
        if percent<0.04:
             model_filename = "/tmp/model_zeros.h5"
             print("Saving model to", model_filename)
             self.model.save(model_filename)





def main():
    # iterator = yield_batch_of_examples("train", 10)
    # test_batch = next(iterator)
    # print(test_batch[0].shape)
    # print(test_batch[1].shape)
    # raise ValueError()

    training_input_helper = training_input.TrainingInput(args.shard_dir)

    config = {
            "alphabet": training_input.input.alphabet,
            "all_lineages": training_input_helper.input_helper.all_lineages,
            "seq_length": 29891,
            "mode": "pruning_style"
        }
    learning_rate = 1e-3
    if not args.starting_model:

        model = modelling.build_model(config)
    else:
        model = modelling.load_saved_model(args.starting_model)

    if args.do_pruning:
        model = modelling.build_pruning_model(config,model,pruning_params={
    'pruning_schedule':
    tfmot.sparsity.keras.PolynomialDecay(initial_sparsity=0.0,
                                         final_sparsity=0.97,
                                         begin_step=0 * 5,
                                         end_step = 400 * 5)
    })

    modelling.compile_model(model, learning_rate)
    print(model.summary())

    batch_size = 32

    # Callback to save model
    checkpoint = tf.keras.callbacks.ModelCheckpoint(
        filepath=f"{args.checkpoint_path}/checkpoint.h5",
        monitor='val_f1_m',
        verbose=1,
        mode='max',
        save_freq=400 * 5)

    # Check if checkpoint path exists and create it if not:
    if not os.path.exists(args.checkpoint_path):
        os.makedirs(args.checkpoint_path)

    # Define a new callback to log the pruning history



    callbacks=[
           
                tfmot.sparsity.keras.UpdatePruningStep(),
                tfmot.sparsity.keras.PruningSummaries(log_dir="./pruning_logs"),
                checkpoint
            ]

    if args.use_wandb:

        import wandb
        from wandb.keras import WandbCallback

        config = dict(learning_rate=learning_rate,
        do_pruning=args.do_pruning,
        initial_model = args.starting_model,
        checkpoint_path = args.checkpoint_path,

        )
        wandb.init(project="sandslash", notes="", config=config)
        callbacks.append( WandbCallback(generator=training_input_helper.yield_batch_of_examples(
                   "test", batch_size),
                            validation_steps=20,
                           log_weights=True),)
    if args.do_pruning:
        callbacks.append(PrintZeros())

    gen = training_input_helper.yield_batch_of_examples("train", batch_size)




   
    if args.epochs is None:
        args.epochs = 1000000

    model.fit_generator(
        gen,
        steps_per_epoch=400,
        epochs=args.epochs,
        validation_data=training_input_helper.yield_batch_of_examples("test", batch_size),
        validation_steps=50,
        callbacks=callbacks
        )
       

if __name__ == "__main__":
    main()



from . import input
import pandas as pd
import numpy as np
import lzma
import gzip

import random


def random_sample_from_list(the_list, proportion):
    return random.sample(the_list, int(proportion * len(the_list)))


num_shards = 400
all_shards = range(num_shards)
train_shards = list(range(num_shards)) #random_sample_from_list(all_shards, 0.8)
test_shards = random_sample_from_list(all_shards, 0.1)

dropout_prob = 0.1
#ref_dropout_prob = 0.05
import gzip


class TrainingInput(object):
    def __init__(self, path):
        self.path = path
        self.input_helper = input.Input(path)
        self.lineage_cache = {}

    def get_multi_hot_from_lineage(self,lineage):
        multi_hot = np.zeros(len( self.input_helper.all_lineages), dtype=np.float32)

        while True:
            if lineage in self.input_helper.lineage_to_index:
                multi_hot[self.input_helper.lineage_to_index[lineage]] = 1
            if "." in lineage:
                subparts = lineage.split(".")
                lineage = ".".join(subparts[:-1])
                continue
            elif lineage in self.input_helper.aliases and self.input_helper.aliases[lineage] != "":
                lineage = self.input_helper.aliases[lineage]
                continue
            else:
                assert lineage == "A" or lineage == "B"
                break
        return multi_hot



    def yield_examples(self,shards):
        while True:
            for shard_num in shards:
                file = gzip.open(f"{self.path}/shard_{shard_num}.tsv.gz","rt")
                for line in file:
                    epi, seq, lineage = line.strip().split("\t")

                    lineage_numpy = self.get_multi_hot_from_lineage_with_cache(
                        lineage)
                    #print(seq)
                    yield (self.input_helper.string_to_one_hot_numpy(seq,target_length = 29891), lineage_numpy)


    


    def get_multi_hot_from_lineage_with_cache(self,lineage):
        # check if lineage is a list, and raise error if so:

        if lineage in self.lineage_cache:
            return self.lineage_cache[lineage]
        lineage_numpy = self.get_multi_hot_from_lineage(lineage)
        self.lineage_cache[lineage] = lineage_numpy
        return lineage_numpy





    def add_dropout(self,generator):
        while True:
            seq, lineage = next(generator)
            random_values = np.random.random((seq.shape[0], 1))
            seq = np.where(
                random_values > dropout_prob  #+ ref_dropout_prob
                ,
                seq,
                0)

            # seq = np.where(random_values < ref_dropout_prob, ref_numpy, seq)
            yield (seq, lineage)


    def get_typed_examples(self,type):
        if type == "train":
            return self.add_dropout(self.yield_examples(train_shards))
        elif type == "test":
            return self.yield_examples(test_shards)


    def yield_batch_of_examples(self,type, batch_size):
        example_iterator = self.get_typed_examples(type)
        while True:
            batch = [next(example_iterator) for _ in range(batch_size)]
            yield (np.stack([x[0] for x in batch]), np.stack([x[1]
                                                            for x in batch]))






import numpy as np
import pandas as pd
from Bio import SeqIO
from . import cov2_genome
import pkg_resources
import json
import lzma
import zipfile
import tarfile
alphabet = "acgt-"
import gzip
import os
import io

from collections import defaultdict
try:
    from . import helpers_compiled
except ImportError:
    print("Attempting pyximport")
    import pyximport
    pyximport.install()
    from . import helpers_compiled


class Input(object):
    def __init__(self, path):
        self.path = path
        self.aliases = json.load(open(os.path.join(path, "aliases.json")))
        self.character_lookup_table = self.make_character_lookup_table(alphabet)
        self.ref_numpy = self.string_to_one_hot_numpy(cov2_genome.seq.lower()[0:29891])
        self.all_lineages = [x.strip() for x in open(os.path.join(path, "all_lineages.txt"))]
        self.lineage_to_index = {x:i for i, x in enumerate(self.all_lineages)}
        self.aliases['XA'] = 'B.1'
        self.lineage_to_level = {x:self.get_unaliased_lineage(x).count(".") for x in self.all_lineages}

    def string_to_ints(self, string):
        as_array = np.array((string), "c")
        as_numbers = as_array.view(np.int8)
        return as_numbers


    def make_character_lookup_table(self, alphabet):
        character_lookup_table = np.zeros((256, len(alphabet)))
        indices = self.string_to_ints(alphabet)
        for i, x in enumerate(indices):
            character_lookup_table[x][i] = 1

        indices = self.string_to_ints(alphabet.upper())
        for i, x in enumerate(indices):
            character_lookup_table[x][i] = 1
        return character_lookup_table


    


    def string_to_one_hot_numpy(self, string, target_length = None):
        if target_length:
            if len(string)> target_length:
                string = string[0:target_length]
            if len(string) < target_length:
                # pad string to desired length
                string = string + "n" * (target_length - len(string))
            
        as_numbers = self.string_to_ints(string)

        return self.character_lookup_table[as_numbers]


    def extract_numpy_features(self,seq,mask_positions, mask_values):
        to_output = np.zeros(len(mask_positions))
        for index_mask in range(len(mask_positions)):
            if seq[mask_positions[index_mask]] == mask_values[index_mask]:
                to_output[index_mask] = 1
        return to_output




    def masked_iterator(self,iterator, mask):
  
        mask_positions = []
        mask_values = []

        for index_full in mask:
            mask_positions.append(index_full//len(alphabet))
            mask_values.append(alphabet[index_full%len(alphabet)])

        mask_positions = np.array(mask_positions, dtype=np.int32)
        mask_values = np.array(mask_values)
        mask_values = mask_values.view(np.int32)
     
        for seq_id, seq in iterator:
            seq = seq.lower()
            to_output = helpers_compiled.extract_numpy_features(seq.encode("utf-8"), mask_positions, mask_values)
            yield seq_id, to_output




    def get_unaliased_lineage(self, lineage):
        components = lineage.split(".")
        if components[0] in self.aliases:
            return self.aliases[components[0]] + "." + ".".join(components[1:])


    def yield_from_fasta(self, filename, mask=None):
        print(f"Attempting to read from fasta file {filename}")
        if filename.endswith(".gz"):
            handle = gzip.open(filename, "rt")
            print("Using gzip mode")
        elif filename.endswith("tar.xz"):
            tar = tarfile.open(filename, "r:xz")
            members = tar.getmembers()
            # find largest member:
            largest_member = max(members, key=lambda x: x.size)
            handle = tar.extractfile(largest_member)
            handle = io.TextIOWrapper(handle, encoding='windows-1252') 
            print("Using largest file as handle")
            print("Using tarxz mode")
        elif filename.endswith(".xz"):
            handle = lzma.open(filename, "rt")
            print("Using xz mode")
        elif filename.endswith(".zip"):
            print("Using zip mode")
            print(f"Opening {filename} as zip")
            the_zip = zipfile.ZipFile(filename)
            # iterate through all files recursively and find "genomic.fna"
            for member in the_zip.infolist():
                if member.filename.endswith("genomic.fna"):
                    handle = the_zip.open(member)
                    break
                raise ValueError("Could not find genomic.fna in zip file")
        else:
            print("Using text mode")
            handle = open(filename, "rt")
        for record in SeqIO.parse(handle, "fasta"):
            if mask != None:
                raise ValueError
            else:
                yield record.id, str(record.seq)[0:29891]


    def apply_mask_to_seq_iterator(self, seq_iterator, selected_indices):
        for seq_id, seq in seq_iterator:
            masked_seq = ""
            for i in selected_indices:
                masked_seq += seq[i]
            yield seq_id, masked_seq


    def apply_numpy_to_seq_iterator(self, seq_iterator):
        for seq_id, seq in seq_iterator:
            yield seq_id, self.string_to_one_hot_numpy(seq).flatten()


    def apply_mask_to_numpy_iterator(self, seq_iterator, selected_indices):
        for seq_id, flat in seq_iterator:
            masked_flat = flat[selected_indices]
            
            yield seq_id, masked_flat
    def batch_singles(self, iterator, batch_size):
        batch = []
        while True:
            try:
                batch.append(next(iterator))
                if len(batch) == batch_size:
                    yield np.stack(batch)
                    batch = []
            except StopIteration:
                if len(batch) > 0:
                    yield np.stack(batch)
                return



import tensorflow as tf
from tensorflow.keras.layers import Input, Dense, Dropout, Flatten, Conv2D, MaxPooling2D, LocallyConnected1D, LocallyConnected2D

from keras import backend as K
from tensorflow.keras import initializers
from tensorflow import keras
import tensorflow_model_optimization as tfmot
from . import input
import numpy as np
import h5py
import gzip

NUM_INIT_FILTERS = 1000


def recall_m(y_true, y_pred):
    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    possible_positives = K.sum(K.round(K.clip(y_true, 0, 1)))
    recall = true_positives / (possible_positives + K.epsilon())
    return recall


def precision_m(y_true, y_pred):
    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    predicted_positives = K.sum(K.round(K.clip(y_pred, 0, 1)))
    precision = true_positives / (predicted_positives + K.epsilon())
    return precision


def f1_m(y_true, y_pred):
    precision = precision_m(y_true, y_pred)
    recall = recall_m(y_true, y_pred)
    return 2 * ((precision * recall) / (precision + recall + K.epsilon()))




prune_low_magnitude = tfmot.sparsity.keras.prune_low_magnitude




class MultiplyByWeights(keras.layers.Layer):
    def __init__(self, weight_reg_value=0.01, **kwargs):
        super(MultiplyByWeights, self).__init__()
        self.weight_reg_value = weight_reg_value

    def build(self, input_shape):
        self.w = self.add_weight(shape=(input_shape[1]),
                                 initializer=initializers.Ones(),
                                 trainable=True,
                                 regularizer=keras.regularizers.l1(
                                     self.weight_reg_value))
        self.the_input_shape = input_shape

    def get_config(self):
        config = super(MultiplyByWeights, self).get_config()
        config['weight_reg_value'] = self.weight_reg_value
        return config

    def get_prunable_weights(self):
        return self.weights

    def call(self, inputs):

        reshaped_weights = K.reshape(self.w, self.the_input_shape[1:])
        return tf.math.multiply(inputs, reshaped_weights)


def build_model(config):

    if config['mode'] == "pretrained_pruned":
        input = tf.keras.Input(shape=(config['features_length']),
                           dtype=tf.float32,
                           name="input")

    else:
        input = tf.keras.Input(shape=(config['seq_length'],
                                  len(config['alphabet'])),
                           dtype=tf.float32,
                           name="input")


    x = Flatten(name="flatten")(input)
    x = MultiplyByWeights(weight_reg_value=1e-8, name="local_1")(x)


    x = Dense(
        NUM_INIT_FILTERS,
        activation='relu',
        name="initial_dense",
    )(x)
    x = Dense(500, activation='relu', name="second_dense")(x)
    x = Dense(len(config['all_lineages']),
              activation='sigmoid',
              name="output_dense")(x)

    model = tf.keras.Model(inputs=input, outputs=x)
    return model


def build_pruning_model(config, model = None, pruning_params = None):
    if model is None:
        model = build_model(config)

    

    def prune_local(layer):
        if isinstance(layer, MultiplyByWeights):
            return tfmot.sparsity.keras.prune_low_magnitude(
                layer, **pruning_params)
        return layer

    model = tf.keras.models.clone_model(
        model,
        clone_function=prune_local,
    )
    return model


def load_saved_model(filename):
    if filename.endswith(".h5.gz"):
        f = gzip.open(filename, 'rb')
        target = h5py.File(f)
    else:
        target = filename
        

    with tfmot.sparsity.keras.prune_scope():
        model = tf.keras.models.load_model(target,
                                           custom_objects={
                                               "f1_m":
                                               f1_m,
                                               "precision_m":
                                               precision_m,
                                               "recall_m":
                                               recall_m,
                                               "MultiplyByWeights":
                                               MultiplyByWeights
                                           }, compile=False)
        return model


def compile_model(model, learning_rate):
    opt = tf.keras.optimizers.Adam(lr=learning_rate, )
    model.compile(
        optimizer=opt,
        loss='binary_crossentropy',
        metrics=['acc', f1_m, precision_m, recall_m, 'binary_crossentropy'])


def create_pretrained_pruned_model(initial_model):


    masking_weights = initial_model.get_layer(
        "prune_low_magnitude_multiply_by_weights").get_weights()
    remaining_positions = np.where(masking_weights[0] != 0)[0]
    num_lineages = initial_model.get_layer(
            "output_dense").output_shape[1]
    print("Number of lineages: ", num_lineages)
    

    model_config = {
        "alphabet": input.alphabet,
        "all_lineages": list(range(num_lineages)),
        "features_length": remaining_positions.shape[0],
        "mode": "pretrained_pruned"
    }
    new_model = build_model(model_config)

    old_initial_weights = initial_model.get_layer(
        "initial_dense").get_weights()
    #print(old_initial_weights[0].shape)
    assert old_initial_weights[0].shape == (29891 * 5, NUM_INIT_FILTERS)
    filtered = old_initial_weights[0][remaining_positions, :]
    new_initial_weights = old_initial_weights
    new_initial_weights[0] = filtered

    print(initial_model.summary())
    for layer in initial_model.layers:
        print(layer.name)

    old_multipliers = masking_weights
    new_multipliers = old_multipliers
    new_multipliers[0] = old_multipliers[0][remaining_positions]

    print(new_model.summary())


    for layer in new_model.layers:
        if layer.name == "initial_dense":
            layer.set_weights(new_initial_weights)
        elif layer.name == "multiply_by_weights_1":
            layer.set_weights(new_multipliers)
        else:
            layer.set_weights(
                initial_model.get_layer(layer.name).get_weights())
    return new_model, remaining_positions



import argparse
import sys
import json

argparser = argparse.ArgumentParser()
argparser.add_argument("fasta_file",
                       help="Fasta file (must be already aligned to ref)", type=str)

argparser.add_argument("--detailed_predictions",
                       help="Save detailed predictions",
                       action="store_true")
argparser.add_argument("--disable_gpu",
                       help="Don't use the GPU",
                       action="store_true")

argparser.add_argument("--threshold",
                          help="Threshold for predictions",
                            type=float,
                            default=0.5)

argparser.add_argument("--custom_full_model",
                          help="Path to custom full model", type=str)

argparser.add_argument("--chunk_size",
                          help="Chunk size for predictions",
                          type=int, default=1000)

args = argparser.parse_args()


import os
import pandas as pd

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3' 

if(args.disable_gpu):
    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
    print("Using CPU", file=sys.stderr)


from . import modelling
from . import input
import numpy as np
import pkg_resources

print("Welcome to Armadillin", file=sys.stderr)
print("", file=sys.stderr)
input_helper = input.Input(pkg_resources.resource_filename("armadillin", "trained_model"))


def get_result(result):
    positive = np.where(result > args.threshold)[0]
    
    positive_lineages = [input_helper.all_lineages[x] for x in positive]
    positive_values = [result[x] for x in positive]
    levels = [input_helper.lineage_to_level[x] for x in positive_lineages]
    max_level = max(levels)
    
    positive_lineages_at_max = [x for x in positive_lineages if input_helper.lineage_to_level[x] == max_level]
    
    lineage_to_results = dict(zip(positive_lineages, positive_values))
    ordered_by_result = sorted(positive_lineages_at_max, key=lambda x: lineage_to_results[x], reverse=True)
    
    
    details = "" if not args.detailed_predictions else "\t"+", ".join([f"{lineage}:{lineage_to_results[lineage]}" for lineage in positive_lineages])
    return ordered_by_result[0], details

def do_predictions(sequence_names, sequence_numpys):
    batched_numpys = input_helper.batch_singles(iter(sequence_numpys), 32)


    results = model.predict(batched_numpys)

    for i, result in enumerate(results):
        
        lineage, details = get_result(result)
        print(f"{sequence_names[i]}\t{lineage}{details}")



if not args.custom_full_model:
    model = modelling.load_saved_model(pkg_resources.resource_filename(__name__, 'trained_model/model_small.h5'))
    #model, mask = modelling.create_pretrained_pruned_model(model)
    mask = json.load(open(pkg_resources.resource_filename(__name__, 'trained_model/mask_small.json')))

else:
    model = modelling.load_saved_model(args.custom_full_model)
    mask = None



filename = args.fasta_file
input_iterator = input_helper.yield_from_fasta(filename)
input_iterator = input_helper.masked_iterator(input_iterator, mask)




print("Predicting...", file=sys.stderr)

def make_predictions():
    large_batch_seq_names = []
    large_batch_seq_numpys = []
    while True:
        try:
            sequence_name, sequence_numpy = next(input_iterator)
            large_batch_seq_names.append(sequence_name)
            large_batch_seq_numpys.append(sequence_numpy)
            if len(large_batch_seq_names) == args.chunk_size:
                do_predictions(large_batch_seq_names, large_batch_seq_numpys)
                large_batch_seq_names = []
                large_batch_seq_numpys = []
        except StopIteration:
            if len(large_batch_seq_names) > 0:
                do_predictions(large_batch_seq_names, large_batch_seq_numpys)
            break


make_predictions()

def main():
    # Sorry for this hack
    pass



import os
os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
from . import modelling
import json
import numpy as np
# Create argparser to capture input and output files
import argparse
parser = argparse.ArgumentParser(description='Create a small model for training')
parser.add_argument('-i', '--input', help='Input file', required=True)
parser.add_argument('-o', '--output', help='Output file', default=None)
parser.add_argument('-d', '--data', help='Data directory', default=None)



args = parser.parse_args()

if not args.output:
    args.output = "./armadillin/trained_model/"
    print(f"You did not enter an output directory (use -o) so we will write to {args.output}")

if not os.path.exists(args.output):
    os.makedirs(args.output)
    

model = modelling.load_saved_model( args.input )
layer = model.get_layer(
"prune_low_magnitude_multiply_by_weights")
weights = layer.get_weights()[0]
num_zeros = np.sum(weights == 0)
num_non_zeros = np.sum(weights != 0)
print("Number of zeros:", num_zeros)
print("Number of non-zeros:", num_non_zeros)
print("Ratio:", num_zeros / num_non_zeros)

model, mask = modelling.create_pretrained_pruned_model(model)

output_model = f'{args.output}/model_small.h5'
output_mask = f'{args.output}/mask_small.json'

os.system(f"cp {args.data}/all_lineages.txt {args.output}/all_lineages.txt")
os.system(f"cp {args.data}/aliases.json {args.output}/aliases.json")
model.save(output_model)
json.dump(mask.tolist(), open(output_mask, 'wt'))

print(f"Mask size was: {len(mask)}")

print(f"Model saved to {output_model}")
print(f"Mask saved to {output_mask}")
